@@ -1,13 +1,11 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from trytond import backend
-from trytond.model import (ModelView, ModelSQL, ModelSingleton, ValueMixin,
-    fields)
-from trytond.pool import Pool
-from trytond.pyson import Eval, Id
-from trytond.tools.multivalue import migrate_property
+from trytond.model import (
+    ModelSingleton, ModelSQL, ModelView, ValueMixin, fields)
 from trytond.modules.company.model import (
     CompanyMultiValueMixin, CompanyValueMixin)
+from trytond.pool import Pool
+from trytond.pyson import Eval, Id, TimeDelta
 
 purchase_invoice_method = fields.Selection(
     'get_purchase_invoice_method', "Invoice Method")
@@ -36,7 +34,12 @@ class Configuration(
                 ]))
     purchase_invoice_method = fields.MultiValue(purchase_invoice_method)
     get_purchase_invoice_method = get_purchase_methods('invoice_method')
-    purchase_process_after = fields.TimeDelta("Process Purchase after",
+    purchase_process_after = fields.TimeDelta(
+        "Process Purchase after",
+        domain=['OR',
+            ('purchase_process_after', '=', None),
+            ('purchase_process_after', '>=', TimeDelta()),
+            ],
         help="The grace period during which confirmed purchase "
         "can still be reset to draft.\n"
         "Applied only if a worker queue is activated.")
@@ -70,26 +73,7 @@ class ConfigurationSequence(ModelSQL, CompanyValueMixin):
             ('company', 'in', [Eval('company', -1), None]),
             ('sequence_type', '=',
                 Id('purchase', 'sequence_type_purchase')),
-            ],
-        depends=['company'])
-
-    @classmethod
-    def __register__(cls, module_name):
-        exist = backend.TableHandler.table_exist(cls._table)
-
-        super(ConfigurationSequence, cls).__register__(module_name)
-
-        if not exist:
-            cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('purchase_sequence')
-        value_names.append('purchase_sequence')
-        fields.append('company')
-        migrate_property(
-            'purchase.configuration', field_names, cls, value_names,
-            fields=fields)
+            ])
 
     @classmethod
     def default_purchase_sequence(cls):
@@ -107,23 +91,6 @@ class ConfigurationPurchaseMethod(ModelSQL, ValueMixin):
     purchase_invoice_method = purchase_invoice_method
     get_purchase_invoice_method = get_purchase_methods('invoice_method')
 
-    @classmethod
-    def __register__(cls, module_name):
-        exist = backend.TableHandler.table_exist(cls._table)
-
-        super(ConfigurationPurchaseMethod, cls).__register__(module_name)
-
-        if not exist:
-            cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('purchase_invoice_method')
-        value_names.append('purchase_invoice_method')
-        migrate_property(
-            'purchase.configuration', field_names, cls, value_names,
-            fields=fields)
-
     @classmethod
     def default_purchase_invoice_method(cls):
         return 'order'
