@@ -381,7 +381,7 @@ class Purchase(
         company = pattern.get('company')
         if not company:
             company = cls.default_company()
-        if company:
+        if company is not None and company >= 0:
             return Company(company).currency.id
 
     @classmethod
